import os
import re
import yaml
import requests
import urlparse

from charmhelpers.core import hookenv
from charmhelpers.core.services import helpers

import actions


keys_and_salts = ('AUTH_KEY', 'SECURE_AUTH_KEY', 'LOGGED_IN_KEY',
                  'NONCE_KEY', 'AUTH_SALT', 'SECURE_AUTH_SALT',
                  'LOGGED_IN_SALT', 'NONCE_SALT')


def get_secrets_from_leader():
    secrets = hookenv.leader_get()
    insufficient_keys = [
            x for x in keys_and_salts
            if x not in secrets
            or len(secrets[x]) != len(actions.mkpasswd())
            ]
    if not insufficient_keys:
        return secrets

    if not hookenv.is_leader():
        return {}  # Try again when we can get all the keys done.

    hookenv.leader_set({k: actions.mkpasswd() for k in insufficient_keys})


def get_redirects(force=False):
    """Return list of redirects relevant for current context:
      - always send redirects to the website relation
      - if the website relation is not present use redirects
        only if ssl is enabled
    """
    config = hookenv.config()
    hook_name = hookenv.hook_name()
    redirects = [
        {
            "match": "/wp-admin",
            "target": "https://{}/wp-admin".format(config["blog_hostname"])
        },
        {
            "match": "/wp-login.php",
            "target": "https://{}/wp-login.php".format(config["blog_hostname"])
        },
    ]
    if force or re.match('website-relation-(joined|changed)', hook_name):
        return redirects
    elif (re.match('webservice-relation-(joined|changed)', hook_name)
          and config['ssl_enabled']
          and not hookenv.relation_ids('website')):
        return redirects
    return []


def get_vhost_options():
    config = hookenv.config()
    if config.get('vhost_options', False):
        vhost_options = yaml.safe_load(config['vhost_options'])
        return vhost_options
    return None


class PluginRelation(helpers.RelationContext):
    name = 'wordpress-plugin'
    interface = 'wordpress-plugin'

    def get_data(self):
        """ jinja won't allow hyphens in identifiers, so we switch to
        underscores here. """

        super(PluginRelation, self).get_data()
        if not hookenv.relation_ids(self.name):
            return
        self['wordpress_plugin'] = self[self.name]

    def as_dict(self):
        return self or {'wordpress_plugin': []}

    def provide_data(self):
        config = hookenv.config()
        path = os.path.join(config['install_path'], 'wp-content/plugins')
        return {'plugin_directory': path}


class ThemeRelation(helpers.RelationContext):
    name = 'wordpress-theme'
    interface = 'wordpress-theme'

    def provide_data(self):
        config = hookenv.config()
        path = os.path.join(config['install_path'], 'wp-content/themes')
        return {'theme_directory': path}


class MemcachedRelation(helpers.RelationContext):
    name = 'memcached'
    interface = 'memcache'


class WebserviceRelation(helpers.RelationContext):
    name = 'webservice'
    interface = 'webservice'

    def get_proxy(self):
        if object_storage():
            target = "{swift-url}/{bucket}/wp-content/uploads/".format(
                **object_storage())
            return [{"match": "/wp-content/uploads/", "proxy_target": target}]
        else:
            return False

    def get_urls(self):
        config = hookenv.config()
        urls = []
        hostnames = [config["blog_hostname"]]
        if config["additional_hostnames"]:
            hostnames.extend(
                [h.strip() for h in
                    config["additional_hostnames"].split(",")])
        for hostname in hostnames:
            urls.append(urlparse.ParseResult(
                        'http',
                        '{}:{}'.format(hostname, config["port_number"]),
                        '', '', '', ''))
            if config["ssl_enabled"]:
                urls.append(urlparse.ParseResult(
                            'https',
                            '{}:{}'.format(hostname, config["ssl_port_number"]),
                            '', '', '', ''))
        return urls

    def provide_data(self):
        config = hookenv.config()
        services = []
        redirects = get_redirects()
        extra_vhost_options = get_vhost_options()
        proxy = self.get_proxy()
        base_vhost = {"type": "php",
                      "document_root": str(config["install_path"]),
                      "webserver_options": ["mod_rewrite", "mod_headers"],
                      "vhost_options": [{'Header': 'append Vary "Cookie"'}],
                      }
        for url in self.get_urls():
            vhost = base_vhost.copy()
            vhost["url"] = url.geturl()
            if proxy:
                vhost["proxy"] = proxy
            if url.scheme == "http" and redirects:
                vhost["redirects"] = redirects
            if "redirects" in config and config["redirects"]:
                vhost["redirect_match"] = yaml.safe_load(config["redirects"])
            if extra_vhost_options:
                vhost["vhost_options"] = vhost["vhost_options"] + extra_vhost_options
            services.append(vhost)

        return {'services': services}


class NEMRelation(helpers.RelationContext):
    name = 'nrpe-external-master'
    interface = 'nrpe-external-master'
    required_keys = ['nagios_hostname']

    def get_data(self):
        """ jinja won't allow hyphens in identifiers, so we switch to
        underscores here. """

        super(NEMRelation, self).get_data()
        if not hookenv.relation_ids(self.name):
            return
        self['nrpe_external_master'] = self[self.name]


class WebsiteRelation(helpers.RelationContext):
    name = 'website'
    interface = 'http'

    # The Services framework is strict on the signature of the provide_data
    # method, hence the workaround below.
    # Used in notify_website_relation() in hooks/actions.py
    def provide_data(self):
        self._provide_data()

    def _provide_data(self, redirects=False):
        config = hookenv.config()
        services = WebserviceRelation().provide_data()['services']
        match = urlparse.urlunparse(
            ('http',
             '{}:{}'.format(config['blog_hostname'], config['port_number']),
             '', '', '', ''))
        target = urlparse.urlunparse(
            ('http',
             '{}:{}'.format(hookenv.unit_private_ip(), config['port_number']),
             '', '', '', ''))
        proxy_http = [{'match': match, 'proxy_target': target}]
        if config['ssl_enabled']:
            match = urlparse.urlunparse(
                ('https',
                 '{}:{}'.format(config['blog_hostname'], config['ssl_port_number']),
                 '', '', '', ''))
            target = urlparse.urlunparse(
                ('https',
                 '{}:{}'.format(hookenv.unit_private_ip(), config['ssl_port_number']),
                 '', '', '', ''))
            proxy_https = [{'match': match, 'proxy_target': target}]
        else:
            proxy_https = []

        refresh = {}
        if object_storage():
            refresh = [{
                'path_regex': '^wp-content/uploads/',
                'min': 525600,  # 1 year
                'percent': 100,
                'max': 525600,
                'options': [
                    'ignore-reload',
                    'ignore-no-store',
                    'ignore-must-revalidate',
                    'ignore-private',
                    'ignore-auth'
                ]
            }]

        new_services = []
        for service in services:
            if urlparse.urlparse(service['url']).scheme == 'http':
                service['proxy'] = proxy_http
                if 'redirects' not in service and redirects:
                    service['redirects'] = get_redirects(force=True)
            elif urlparse.urlparse(service['url']).scheme == 'https' and proxy_https:
                service['proxy'] = proxy_https
            if refresh:
                service['refresh'] = refresh
            if 'type' in service:
                del(service['type'])
            if 'document_root' in service:
                del(service['document_root'])
            if 'redirect_match' in service:
                del(service['redirect_match'])
            new_services.append(service)
        return {'services': new_services}


# FIXME: this really doesn't belong here, but we're avoiding circular imports
# right now.
def call_wordpress(uri, redirects=True, payload={}, ssl=False, verify_ssl=True):
    config = hookenv.config()
    headers = {'Host': config['blog_hostname']}
    url = urlparse.urlunparse((
        ("https" if ssl else "http"),
        'localhost:{}'.format(
            (config["ssl_port_number"] if ssl else config['port_number'])),
        uri, '', '', ''))
    if payload:
        return requests.post(url, allow_redirects=redirects,
                             headers=headers, data=payload,
                             verify=verify_ssl)
    else:
        return requests.get(url, allow_redirects=redirects,
                            headers=headers, verify=verify_ssl)


class wordpress_configured(dict):
    """Check whether first install has been completed."""

    def __bool__(self):
        # Check if we have WP code deployed at all
        if not vhost_ready():
            return False
        # We have code on disk, check if configured
        try:
            r = call_wordpress('/', redirects=False)
        except requests.exceptions.ConnectionError:
            return False
        if (r.status_code == 302 and
            re.match('^.*/wp-admin/install.php',
                     r.headers.get('location', ''))):
            return False
        else:
            return True

    def __nonzero__(self):
        return self.__bool__()

    def is_ready(self):
        return self.__bool__()


class vhost_ready(dict):
    """Check whether wordpress is available using http."""

    def __bool__(self):
        # Check if we have WP code deployed at all
        try:
            r = call_wordpress('/wp-login.php', redirects=False)
        except requests.exceptions.ConnectionError:
            return False
        if r.status_code in (403, 404):
            return False
        else:
            return True

    def __nonzero__(self):
        return self.__bool__()

    def is_ready(self):
        return self.__bool__()


class object_storage(dict):
    def __init__(self):
        if wordpress_configured():
            object_storage_opts = actions.get_option('object_storage')
        else:
            return

        if (object_storage_opts
                and 'swift-url' in object_storage_opts
                and 'bucket' in object_storage_opts):
            self['swift-url'] = object_storage_opts['swift-url']
            self['bucket'] = object_storage_opts['bucket']
        else:
            return
